import concurrent.futures
import contextlib
import ipaddress
import json
import logging
import os
import random
import socket
import subprocess
import tempfile
import time
from pathlib import Path

import orchestrator
import pytest
import yaml
from alembic import command
from alembic.config import Config
from faker import Faker
from faker.providers import BaseProvider
from orchestrator import app_settings
from orchestrator.db import Database, db
from orchestrator.db.database import ENGINE_ARGUMENTS, SESSION_ARGUMENTS, BaseModel
from sqlalchemy import create_engine
from sqlalchemy.engine import make_url
from sqlalchemy.orm import scoped_session, sessionmaker
from starlette.testclient import TestClient

from gso.main import init_gso_app

logger = logging.getLogger(__name__)

CONTAINER_UP_TIMEOUT_S = 40
CONTAINER_HEALTH_TIMEOUT_S = 60


def _wait_for_container_to_be_healthy(container_name):
    def _is_container_healthy():
        args = ["docker", "inspect", container_name]
        # just die badly in case of any problems
        # (e.g. json decoding error, no health in output)
        inspect_output = json.loads(_run_command_or_fail(args))

        # barf if this element isn't the output
        # ... we require a health check in the container
        return inspect_output[0]["State"]["Health"]["Status"] == "healthy"

    start = time.time()
    while True:
        if time.time() - start > CONTAINER_HEALTH_TIMEOUT_S:
            break
        if _is_container_healthy():
            return True
        time.sleep(1)

    return False


def _run_command_or_fail(args):
    dc = subprocess.Popen(args, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, encoding="utf-8")  # noqa S603
    output_info, _ = dc.communicate(timeout=CONTAINER_UP_TIMEOUT_S)
    assert dc.returncode == 0, output_info
    return output_info


def _use_isolated_docker_compose():
    if not bool(os.environ.get("USE_ISOLATED_DOCKER_COMPOSE")):
        logger.warning('"USE_ISOLATED_DOCKER_COMPOSE" is not defined in the environment')
        return False

    def _is_docker_compose_installed():
        compose_result = subprocess.call(
            ["docker", "compose"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL  # noqa S603
        )

        if compose_result == 0:
            return True

        compose_result = subprocess.call(
            ["docker-compose"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL  # noqa S603
        )

        return compose_result == 0

    assert _is_docker_compose_installed(), "docker-compose is not available in the environment"
    return True


@contextlib.contextmanager
def run_docker_compose(compose_filename, container_names):
    # make a random project name, rather than some env dirname
    project_name = f"test-{random.randint(1000, 10000)}"  # noqa S311

    def _get_compose_command_arguments(params):
        compose_result = subprocess.call(
            ["docker", "compose"], stdout=subprocess.DEVNULL, stderr=subprocess.DEVNULL  # noqa S603
        )
        if compose_result:
            return ["docker-compose"] + params
        return ["docker", "compose"] + params

    args = _get_compose_command_arguments(["-f", compose_filename, "-p", project_name, "up", "--detach"])

    _run_command_or_fail(args)

    try:
        with concurrent.futures.ThreadPoolExecutor(max_workers=2) as executor:
            future_to_container_name = {
                executor.submit(_wait_for_container_to_be_healthy, name): name for name in container_names
            }

            for future in concurrent.futures.as_completed(future_to_container_name):
                name = future_to_container_name[future]
                assert future.result(), f"health check failed for container {name}"

            yield  # wait here until the context finishes

    finally:
        for name in container_names:
            try:
                args = ["docker", "logs", name]
                logging.info(_run_command_or_fail(args))
            except:  # noqa: E722
                # crappy bare except, but this is debugging code
                # continue with cleanup - but log an error message about this
                logging.exception(f"error calling `docker logs {name}`")

        args = _get_compose_command_arguments(["-f", compose_filename, "-p", project_name, "down"])
        _run_command_or_fail(args)


@pytest.fixture(scope="session")
def free_host_ports():
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as postgres:
        postgres.bind(("", 0))
        postgres.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        return {"postgres": postgres.getsockname()[1]}


@pytest.fixture(scope="session")
def postgres_db_params(free_host_ports):
    return {
        "hostname": "localhost",
        "port": free_host_ports["postgres"],
        "username": "nwa",
        "password": "nwa",
        "database": "gso-test-db",
    }


@pytest.fixture(scope="session")
def postgres_container_params(free_host_ports, postgres_db_params):
    container_name = f"gso-postgres-{random.randint(1000, 10000)}"  # noqa S311

    yield {
        "image": "postgres:15.4",
        "container_name": container_name,
        "ports": [f'{free_host_ports["postgres"]}:5432'],
        "environment": {
            "POSTGRES_USER": postgres_db_params["username"],
            "POSTGRES_PASSWORD": postgres_db_params["password"],
            "POSTGRES_DB": postgres_db_params["database"],
        },
        "healthcheck": {
            "interval": "2s",
            "timeout": "2s",
            "retries": 3,
            "test": [
                "CMD-SHELL",
                f"pg_isready -U {postgres_db_params['username']} -d {postgres_db_params['database']}",
            ],
        },
    }


@pytest.fixture(scope="session")
def testenv_docker_compose_params(postgres_container_params):
    compose_config = {
        "version": "3.1",
        "services": {
            "gso-postgres-server": postgres_container_params,
        },
    }
    with tempfile.NamedTemporaryFile(mode="w", suffix=".yaml") as f:
        yaml.dump(compose_config, f)
        f.flush()

        yield {
            "compose_filename": f.name,
            "containers": [service["container_name"] for service in compose_config["services"].values()],
        }


@pytest.fixture(scope="session")
def testenv_docker_containers(testenv_docker_compose_params):
    if _use_isolated_docker_compose():
        with run_docker_compose(
            compose_filename=testenv_docker_compose_params["compose_filename"],
            container_names=testenv_docker_compose_params["containers"],
        ):
            # wait here until the caller context finishes
            yield
    else:
        yield


class FakerProvider(BaseProvider):
    def ipv4_network(self):
        ipv4 = self.generator.ipv4()
        interface = ipaddress.IPv4Interface(ipv4 + "/24")
        network = interface.network.network_address

        return ipaddress.IPv4Network(str(network) + "/24")

    def ipv6_network(self):
        ipv6 = self.generator.ipv6()
        interface = ipaddress.IPv6Interface(ipv6 + "/64")
        network = interface.network.network_address

        return ipaddress.IPv6Network(str(network) + "/64")


@pytest.fixture(scope="session")
def faker() -> Faker:
    fake = Faker()
    fake.add_provider(FakerProvider)
    return fake


@pytest.fixture(scope="session")
def configuration_data() -> dict:
    with contextlib.closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as s:
        s.bind(("", 0))
        s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        yield {
            "GENERAL": {"public_hostname": "https://gap.geant.org"},
            "RESOURCE_MANAGEMENT": {"todo": "todo"},
            "IPAM": {
                "INFOBLOX": {
                    "scheme": "https",
                    "wapi_version": "2.12",
                    "host": "10.0.0.1",
                    "username": "robot-user",
                    "password": "robot-user-password",
                },
                "LO": {
                    "V4": {"containers": [], "networks": ["10.255.255.0/26"], "mask": 32},
                    "V6": {"containers": [], "networks": ["dead:beef::/80"], "mask": 128},
                    "domain_name": ".lo",
                    "dns_view": "default",
                },
                "TRUNK": {
                    "V4": {"containers": ["10.255.255.0/24", "10.255.254.0/24"], "networks": [], "mask": 31},
                    "V6": {"containers": ["dead:beef::/64", "dead:beee::/64"], "networks": [], "mask": 126},
                    "domain_name": ".trunk",
                    "dns_view": "default",
                },
                "GEANT_IP": {
                    "V4": {"containers": ["10.255.255.0/24", "10.255.254.0/24"], "networks": [], "mask": 31},
                    "V6": {"containers": ["dead:beef::/64", "dead:beee::/64"], "networks": [], "mask": 126},
                    "domain_name": ".geantip",
                    "dns_view": "default",
                },
                "SI": {
                    "V4": {"containers": ["10.255.253.128/25"], "networks": [], "mask": 31},
                    "V6": {"containers": [], "networks": [], "mask": 126},
                    "domain_name": ".geantip",
                    "dns_view": "default",
                },
                "LT_IAS": {
                    "V4": {"containers": ["10.255.255.0/24"], "networks": [], "mask": 31},
                    "V6": {"containers": ["dead:beef:cc::/48"], "networks": [], "mask": 126},
                    "domain_name": ".geantip",
                    "dns_view": "default",
                },
            },
            "PROVISIONING_PROXY": {
                "scheme": "https",
                "api_base": "localhost:44444",
                "auth": "Bearer <token>",
                "api_version": 1123,
            },
        }


@pytest.fixture(scope="session")
def data_config_filename(configuration_data) -> str:
    file_name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
    open(file_name, "x").close()
    with open(file_name, "wb") as f:
        f.write(json.dumps(configuration_data).encode("utf-8"))
        f.flush()

        os.environ["OSS_PARAMS_FILENAME"] = f.name

        yield f.name


@pytest.fixture(scope="session")
def db_uri(postgres_db_params):
    """Provide the database uri configuration to run the migration on."""
    if _use_isolated_docker_compose():
        return (
            f"postgresql://{postgres_db_params['username']}:{postgres_db_params['password']}"
            f"@{postgres_db_params['hostname']}:{postgres_db_params['port']}/{postgres_db_params['database']}"
        )

    return os.environ.get("DATABASE_URI_TEST", "postgresql://nwa:nwa@localhost/gso-test-db")


def run_migrations(db_uri: str) -> None:
    """Configure the alembic migration and run the migration on the database.

    Args:
    ----
    db_uri: The database uri configuration to run the migration on.

    Returns:
    -------
    None
    """

    path = Path(__file__).resolve().parent
    app_settings.DATABASE_URI = db_uri
    alembic_cfg = Config(file_=path / "../gso/alembic.ini")
    alembic_cfg.set_main_option("sqlalchemy.url", db_uri)

    alembic_cfg.set_main_option("script_location", str(path / "../gso/migrations"))
    version_locations = alembic_cfg.get_main_option("version_locations")
    alembic_cfg.set_main_option(
        "version_locations", f"{version_locations} {os.path.dirname(orchestrator.__file__)}/migrations/versions/schema"
    )

    command.upgrade(alembic_cfg, "heads")


@pytest.fixture(scope="session")
def database(db_uri, testenv_docker_containers):
    """Create database and run migrations and cleanup after wards.

    Args:
    ----
    db_uri: The database uri configuration to run the migration on.
    testenv_docker_containers: The docker containers to run the migration on.
    """

    db.update(Database(db_uri))
    url = make_url(db_uri)
    db_to_create = url.database
    url = url.set(database="postgres")

    engine = create_engine(url)
    with engine.connect() as conn:
        conn.execute("COMMIT;")
        conn.execute(f"SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='{db_to_create}';")  # noqa
        conn.execute(f'DROP DATABASE IF EXISTS "{db_to_create}";')  # noqa
        conn.execute("COMMIT;")
        conn.execute(f'CREATE DATABASE "{db_to_create}";')  # noqa

    run_migrations(db_uri)
    db.wrapped_database.engine = create_engine(db_uri, **ENGINE_ARGUMENTS)

    try:
        yield
    finally:
        db.wrapped_database.engine.dispose()
        with engine.connect() as conn:
            conn.execute("COMMIT;")
            # Terminate all connections to the database
            conn.execute(
                f"SELECT pg_terminate_backend(pid) FROM pg_stat_activity WHERE datname='{db_to_create}';"  # noqa
            )
            conn.execute(f'DROP DATABASE IF EXISTS "{db_to_create}";')  # noqa


@pytest.fixture(autouse=True)
def db_session(database):
    """Ensure that tests are executed within a transactional scope that automatically rolls back after completion.

    This fixture facilitates a pattern known as 'transactional tests'. At the start, it establishes a connection and
    begins an overarching transaction. Any database operations performed within the test function—whether they commit
    or not happen within the context of this master transaction.

    From the perspective of the test function, it seems as though changes are getting committed to the database,
    enabling the tests to query and assert the persistence of data. Yet, once the test completes, this fixture
    intervenes to roll back the master transaction. This ensures a clean slate after each test, preventing tests from
    polluting the database state for subsequent tests.

    Benefits:
    - Each test runs in isolation with a pristine database state.
    - Avoids the overhead of recreating the database schema or re-seeding data between tests.

    Args:
    ----
    database: A fixture reference that initializes the database.
    """

    with contextlib.closing(db.wrapped_database.engine.connect()) as test_connection:
        # Create a new session factory for this context.
        session_factory = sessionmaker(bind=test_connection, **SESSION_ARGUMENTS)
        scoped_session_instance = scoped_session(session_factory, scopefunc=db.wrapped_database._scopefunc)

        # Point the database session to this new scoped session.
        db.wrapped_database.session_factory = session_factory
        db.wrapped_database.scoped_session = scoped_session_instance

        # Set the query for the base model.
        BaseModel.set_query(scoped_session_instance.query_property())
        transaction = test_connection.begin()
        try:
            yield
        finally:
            transaction.rollback()
            scoped_session_instance.remove()


@pytest.fixture(scope="session", autouse=True)
def fastapi_app(database, db_uri):
    """Load the GSO FastAPI app for testing purposes."""

    app_settings.DATABASE_URI = db_uri
    return init_gso_app(settings=app_settings)


@pytest.fixture(scope="session")
def test_client(fastapi_app):
    return TestClient(fastapi_app)
